VEILID_SERVICE_URL = "http://veilid:80"
# Service name of our traefik service
# TODO: Remove this once when we remove reverse proxy in Veilid Connection
VEILID_SYFT_PROXY_URL = "http://proxy:80"
HEALTHCHECK_ENDPOINT = "/healthcheck"
GEN_VLD_KEY_ENDPOINT = "/generate_vld_key"
RET_VLD_KEY_ENDPOINT = "/retrieve_vld_key"
VEILID_PROXY_PATH = "/proxy"
